#include <stdlib.h>

#include "ddplot.h"
#include "messages.h"
#include "neighlist.h"
#include "pltwin.h"
#include "rdf.h"


/*--------------------------------------------------------------------------------
!
!  This function finds the nearest neighbors of atoms up to the cutoff distance
!  specified by maxdist. The PLT files generated by BOP do not give periodicity
!  in x, y and z direction but instead the size of the block in these directions.
!  However, in some PLT files the three numbers refer to periodicity. 
!  ALL .PLT FILES SHOULD GIVE PERIODICITY ALONG X,Y,Z DIRECTIONS.
!
!-------------------------------------------------------------------------------*/

void BuildNeighborList( PltWin *pw, int nnlist[MAX_ATOMS][MAX_NEIGHBORS] )
{
  FILE *f;
  double dist, dx, dy, dz, maxdist, nndist;
  int icell[3], icmin[3], icmax[3];
  int icen, ineigh, idx, i, d, nn, iidx, jidx, kidx;

 
  // Calculate the cutoff radius for seeking atomic neighbors. This number is used
  // to construct the linked neighbor list. If each cell is at least rcut x rcut x rcut,
  // the neighbors can be searched only in the adjacent cells.
  if (NEIGHRCUT>0)  // given in .ddplot by user
    pw->rcut = NEIGHRCUT;
  else {  // no given value -> try to guess it -> may result in the error message if too large
    if (!pw->isGBfile) 
      pw->rcut = sqrt(pow(pw->period[0],2) + pow(pw->period[1],2) + pow(pw->period[2],2));
    else
      pw->rcut = pw->blSize[0]/10.0;  // this is only a guess - can be changed as we like
  }

  if (VERBOSE) {
    printf("* constructing the linked list of atomic neighbors\n");
    printf("    NeighRCut = %7.4lf\n", pw->rcut);
    printf("  (if this cut-off radius is too large, ddplot says it exceeded MAX_NEIGHBORS ->\n");
    printf("   set NeighRCut manually in the .ddplot file)\n");
  }


  // calculate the radial distribution function
  NeighInit(pw);
  NeighList(pw);
  CalcRDF(pw);


  // return the nearest neighbors
  for (icen=0; icen<pw->NInit; icen++) {
    // consider only atoms in the selected Z-layers
    if (!pw->zLayerSel[ pw->zLayer[icen] ])
      continue;

    idx = 0;
  
    // get the indices of the cell in which atom icen resides
    GetCell(pw, icen, icell);

    // we will search also the neighboring cells - get the indices
    for (d=0; d<3; d++) {
      icmin[d] = (icell[d]-1 >= 0)  ?  icell[d]-1 : 0;  
      icmax[d] = (icell[d]+1 <= pw->ncell[d]-1)  ?  icell[d]+1 : pw->ncell[d]-1;
    }


    // search through all neighboring cells and itself
    for (iidx=icmin[0]; iidx<=icmax[0]; iidx++) {
      for (jidx=icmin[1]; jidx<=icmax[1]; jidx++) {
	for (kidx=icmin[2]; kidx<=icmax[2]; kidx++) {

	  ineigh = pw->head[iidx][jidx][kidx];

	  while (ineigh>=0) {
	    // skip itself
	    if (ineigh==icen) {
	      ineigh = pw->list[ineigh];
	      continue;
	    }

	    // consider only the neighboring atoms located in selected Z-layers
	    if (!pw->zLayerSel[ pw->zLayer[ineigh] ]) {
	      ineigh = pw->list[ineigh];
	      continue;
	    }
	    
	    dx = pw->xyzInit[ineigh][0] - pw->xyzInit[icen][0];
	    dy = pw->xyzInit[ineigh][1] - pw->xyzInit[icen][1]; 
	    dz = pw->xyzInit[ineigh][2] - pw->xyzInit[icen][2]; 
	    
	    // calculate the distance between icen and ineigh
	    dist = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2));

	    // check the neighbor distances (arrNeighbors says how many
	    // shortest neighbor distances are regarded for plotting)      
	    if (dist<=pw->rcut) {
	      for (nn=0; nn<pw->arrNeighNum; nn++) {
		nndist = pw->RdfDist[ pw->arrNeighbors[nn] - 1 ];
		
		if (dist<=nndist+1e-3 && dist>=nndist-1e-3) {
		  if (idx>=MAX_NEIGHBORS) {
		    msgCritical( crit_MAX_NEIGHBORS_Exceeded );
		    exit(0);	    
		  }
		  
		  nnlist[icen][idx] = ineigh;
		  idx++;
		  break;
		}
	      }
	    }
	    
	    ineigh = pw->list[ineigh];
	  }
	}
      }
    }

    nnlist[icen][idx] = END_OF_LIST;
  }


  if (VERBOSE) {
    printf("* calculated radial distribution function (distance, number of atomic pairs):\n");
    for (i=0; i<pw->RdfNNeigh; i++)
      printf("    %7.4lf  %3d\n", pw->RdfDist[i], pw->RdfNum[i]);
  }
}


/*---------------------------------------------------------------------------
!
!  Calculation of the cell size and the number of cells that fill the
!  block so that a neighbor of an atom will not be searched farther than
!  in the immediately adjacent cells.
!
!--------------------------------------------------------------------------*/

void NeighInit( PltWin *pw )
{
  int d;

  for (d=0; d<3; d++) {
    // number of cells in (x,y,z) directions
    pw->ncell[d] = 1 + trunc(pw->blSize[d] / (1.001*pw->rcut));

    if (pw->ncell[d] >= MAX_CELL) {
      msgCritical( crit_MAXCELL_Exceeded );
      exit(0);
    }

    // size of each cell's edge in (x,y,z) directions
    pw->cellsize[d] = pw->blSize[d] / pw->ncell[d];
  }
}


/*---------------------------------------------------------------------------
!
!  Construct a linked neighbor list. Whole block is divided into separate
!  cells whose width is larger than the cut-off radius rcut. The number of
!  cells along each axis is stored in <ncell>. First atom (any atom) in each
!  cell is stored as <head>. Then, <list> contains a chain of other atoms in
!  the same cell. For example, if <head> equals to 15, go to list(15) to
!  find the next atom. If list(15)=3, then go to list(3) and so on. If the
!  list(#) gives -1, we are at the end of the list for this cell.
!
---------------------------------------------------------------------------*/

void NeighList( PltWin *pw )
{
  int icell[3], i, iidx, jidx, kidx, j, k;

  // initially all -1
  memset(pw->head, -1, sizeof(pw->head));

  for (i=0; i<pw->NInit; i++) {
    // get the indices of the cell in which atom <i> lies
    GetCell( pw, i, icell );

    iidx = icell[0];
    jidx = icell[1];
    kidx = icell[2];

    pw->list[i] = pw->head[iidx][jidx][kidx];   // first in the list is -1 (terminator)
    pw->head[iidx][jidx][kidx] = i;             // head is the last atom
  }
}


/*---------------------------------------------------------------------------
!
!  Gives the indices of the cell in which atom <i> lies. For example,
!  icell=(0,5,3) means that the atom is located in the zeroth column along
!  the x-direction, the fifth row along the y-direction and the third layer
!  along the z-direction.
!
!--------------------------------------------------------------------------*/

void GetCell( PltWin *pw, int i, int icell[3] )
{
  int d;

  for (d=0; d<3; d++) {
    icell[d] = (int) trunc( (pw->xyzInit[i][d]-pw->xyzMin[d]) / pw->cellsize[d]);

    // in exceptional cases it can happen that xyzInit[i][d] is exactly at the cell
    // edge -> then icell[d] = ncell[d] and we need to subtract 1 to get the
    // correct cell number
    if (icell[d] == pw->ncell[d] && pw->xyzInit[i][d]-pw->xyzMin[d]-pw->blSize[d] < 1e-6) 
      icell[d]--;

    if (icell[d] >= pw->ncell[d] || icell[d] < 0) {
      msgError( err_AtomNotInCell );      
      exit(0);
    }
  }
}
