#include <math.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <stdlib.h>

#include "calcs.h"
#include "messages.h"
#include "misc.h"
#include "neighlist.h"
#include "pltwin.h"




// Finds the atom in the topmost layer that is nearest to the position [x,y] on the screen
int PltWin::FindNearestAtom( int x, int y )
{
  double dist,mindist;
  int n,atom;

  atom = -1;
  for (n=0; n<NInit; n++) {
    dist = sqrt( pow(atomScreen[n][0]-x,2) + pow(atomScreen[n][1]-y,2) );

    if (dist<mindist || atom<0) {
      mindist = dist;
      atom = n;
    }
  }

  return atom;
}


//-----------------------------------------------------------------------------
//
//   Reading the BOP block file (.bl, .block)
//
//-----------------------------------------------------------------------------

bool PltWin::LoadBOP( QString fname )
{
  const char *fn = fname.ascii();
  FILE *f;
  QString rec;
  char *buf;
  QString elem;
  double latpar;
  int n, idx, d;
  bool eof, ok=true;

  FName = fname;
  f = fopen(fn,"r+");
  if (f==NULL)
    return false;

  buf = (char *) malloc(128);

  // reading the plot file
  while (!feof(f)) {
    eof = fgets(buf,128,f)==NULL;
    if (eof)
      break;

    // throw out comments
    rec = buf;
    idx = rec.find('#',0);   rec.truncate(idx);
    idx = rec.find('\n',0);  rec.truncate(idx);
    idx = rec.find('\r',0);  rec.truncate(idx);  // Linux terminates with \r\n

    // length of the simulation cell (in units of the lattice parameter)
    if (rec=="LEN") {
      fscanf(f, "%lf %lf %lf\n", &period[0], &period[1], &period[2]);
      continue;
    }

    // lattice parameter
    if (rec=="LATPAR") {
      fscanf(f, "%lf\n", &latpar);
      continue;
    }

    // relaxed coordinates of atoms
    if (rec=="ND") { 
      for (d=0; d<3; d++)
	period[d] /= latpar;  // the correct periodicity along the three directions
      lenBurgers = period[2];   // for screw dislocations

      fscanf(f, "%d\n", &NRel);  
      fgets(buf,128,f);  // skip D token

      if (NRel>MAX_ATOMS) {
	msgCritical( crit_MAX_ATOMS_Exceeded );
	ok = false;
	break;
      }

      for (n=0; n<NRel; n++) {
	fscanf(f, "%lf %lf %lf", &xyzRel[n][0], &xyzRel[n][1], &xyzRel[n][2]);
	fgets(buf,128,f);  // skip the rest, if there is some junk

	for (d=0; d<3; d++)
	  xyzRel[n][d] *= period[d];  // correct positions of relaxed atoms
      }
      continue;
    }   

    // inert atoms near the surface of the simulated block
    if (rec=="NINERT") {
      fscanf(f, "%d\n", &NInert);
      fgets(buf,128,f);  // skip DINERT token

      if (NInert>MAX_ATOMS) {
	msgCritical( crit_MAX_ATOMS_Exceeded );
	ok = false;
	break;
      }

      for (n=0; n<NInert; n++) {
	fscanf(f, "%lf %lf %lf", &xyzInert[n][0], &xyzInert[n][1], &xyzInert[n][2]);
	fgets(buf,128,f);  // skip the rest, if there is some junk

	for (d=0; d<3; d++)
	  xyzInert[n][d] *= period[d];  // correct positions of inert atoms
      }
      continue;
    }

     // unrelaxed coordinates of atoms
    if (rec=="UNRLD") {
      NInit = NRel;  // NRel comes from block under D
      for (n=0; n<NInit; n++) {
	fscanf(f, "%lf %lf %lf", &xyzInit[n][0], &xyzInit[n][1], &xyzInit[n][2]);
	fgets(buf,128,f);  // skip the rest, if there is some junk
      }
      continue;
     }
  }

  free(buf);
  fclose(f);

  return ok;
}


//-----------------------------------------------------------------------------
//
//   Reading the block in the DDPlot internal format (.dd)
//
//-----------------------------------------------------------------------------

bool PltWin::LoadDD( QString fname )
{
  const char *fn = fname.ascii();
  FILE *f;
  QString cmd;
  char *buf;
  long int pos;
  int i,idx;
  bool eof, ok=true;

  FName = fname;
  f = fopen(fn,"r+");
  if (f==NULL)
    return false;

  buf = (char *) malloc(128);

  // reading the plot file
  while (!feof(f)) {
    eof = fgets(buf,128,f)==NULL;
    if (eof)
      break;

    // throw out comments
    cmd = buf;
    idx = cmd.find('#',0);   cmd.truncate(idx);
    idx = cmd.find('\n',0);  cmd.truncate(idx);
    idx = cmd.find('\r',0);  cmd.truncate(idx);  // Linux terminates with \r\n
    if (cmd.isEmpty())
      continue;

    cmd = cmd.upper();

    // interpretation of commands    
    if (cmd=="CSYS") {
      for (i=0; i<3; i++)
	fscanf(f, "%lf %lf %lf\n", &csys[i][0], &csys[i][1], &csys[i][2]);
      continue;
    }

    if (cmd=="PERIOD") {
      fscanf(f, "%lf %lf %lf\n", &period[0], &period[1], &period[2]);
      lenBurgers = period[2];   // for screw dislocations
      continue;
    }

    if (cmd=="DISLO_CENTER") {
      fscanf(f, "%lf %lf\n", &xCore, &yCore);
      continue;
    }

    if (cmd=="NUM_UNREL") {
      fscanf(f,"%d\n",&NInit);
      if (NInit>MAX_ATOMS) {
	msgCritical( crit_MAX_ATOMS_Exceeded );
	ok = false;
	break;
      }
      continue;
    }

    if (cmd=="COOR_UNREL") {
      for (i=0; i<NInit; i++)
	fscanf(f, "%lf %lf %lf %d\n", 
	       &xyzInit[i][0], &xyzInit[i][1], &xyzInit[i][2], &atomType[i]);
      continue;
    }

    if (cmd=="NUM_REL") {
      fscanf(f,"%d\n",&NRel);
      if (NRel>MAX_ATOMS) {
	msgCritical( crit_MAX_ATOMS_Exceeded );
	ok = false;
	break;
      }
      continue;
    }

    if (cmd=="COOR_REL") {
      for (i=0; i<NRel; i++)
	fscanf(f, "%lf %lf %lf\n", 
	       &xyzRel[i][0], &xyzRel[i][1], &xyzRel[i][2]);
      continue;
    }

    // read the rest of the file - information about the relaxation
    if (cmd=="RELAX_DETAILS") {
      pos = ftell(f);
      fseek(f,0,SEEK_END);
      TailSize = ftell(f)-pos+1;
      fseek(f,pos,SEEK_SET);
      
      tail = (char*) malloc(TailSize);
      if (tail==NULL) {
      	msgCritical(crit_OutOfMemory);
      	exit(2);
      }
      
      fread(tail,1,TailSize,f);
    }

  }

  free(buf);
  fclose(f);

  return ok;
}


/*--------------------------------------------------------------------

  Reading the block containing the grain boundary data file (.PLT
  format).  The first configuration in the block is regarded as the
  unrelaxed (starting) configuration. The arrows then mean the
  relative displacements between an actual atomic positions and the
  atomic positions in this unrelaxed configuration.

--------------------------------------------------------------------*/

bool PltWin::LoadGB( QString fname, long int *blstart, PltWin *initConf )
{
  FILE *f;
  char *buf, c;
  char txt[80];
  int dummy, endflag;
  int n, d;


  // the file read is the grain boundary data file
  isGBfile = true;
  FName = fname;

  f = fopen(fname.ascii(),"r+");
  if (f==NULL)
    return false;

  // set the position to the start of a given block
  fseek(f, *blstart, SEEK_SET);

  // read one block
  fscanf(f,"%d\n",&NRel);

  if (NRel>MAX_ATOMS) {
    msgCritical( crit_MAX_ATOMS_Exceeded );
    fclose(f);
    return false;
  }
  
  // relaxed block - reading coordinates
  for (n=0; n<NRel; n++)
    fscanf(f, "%lf\n", &xyzRel[n][2]);
  
  for (n=0; n<NRel; n++) {
    fscanf(f, "%lf %lf %d %d %d\n", 
	   &xyzRel[n][0], &xyzRel[n][1], &atomType[n], &gbLayer[n], &dummy);
    if (gbLayer[n]==0)  gbYcoor = xyzRel[n][1];
  }


  // Initial configuration is identical to the relaxed configuration only
  // for the first data set in the file. For all other sets, the unrelaxed
  // configuration is taken from the first block in the file.

  if (*blstart==0) {  
    NInit = NRel;
    for (n=0; n<NInit; n++) 
      for (d=0; d<3; d++)
	xyzInit[n][d] = xyzRel[n][d];    
  } else {
    NInit = initConf->NInit;
    for (n=0; n<initConf->NInit; n++) 
      for (d=0; d<3; d++)
	xyzInit[n][d] = initConf->xyzInit[n][d];  
  }
  
  // skip the trailing zero - what if it is nonzero ?
  fscanf(f, "%d\n", &endflag);  

  // save the position of the start of the subsequent block; if this is the last one,
  // return -1
  if (feof(f))
    *blstart = -1;
  else
    *blstart = ftell(f);

  fclose(f);
  return true;
}


//-----------------------------------------------------------------------------
//
//   Reading the block in the plot format of BOP (.plt)
//
//-----------------------------------------------------------------------------

bool PltWin::LoadPLT( QString fname )
{
  const char *fn = fname.ascii(); 
  FILE *f;
  QString qtxt;
  char *buf, c;
  char txt[80];
  double unused;
  long int pos;
  int n, d;

  FName = fname;
  f = fopen(fn,"r+");
  if (f==NULL)
    return false;

  //
  // relaxed block - reading coordinates
  //

  fscanf(f,"%d\n",&NRel);

  if (NRel>MAX_ATOMS) {
    msgCritical( crit_MAX_ATOMS_Exceeded );
    fclose(f);
    return false;
  }

  for (n=0; n<NRel; n++)
    fscanf(f, "%lf\n", &xyzRel[n][2]);

  for (n=0; n<NRel; n++)
    fscanf(f, "%lf %lf %d\n", &xyzRel[n][0], &xyzRel[n][1], &atomType[n]);


  //
  // initial block - reading coordinates
  //

  fscanf(f,"%d\n",&NInit);

  if (NInit>MAX_ATOMS) {
    msgCritical( crit_MAX_ATOMS_Exceeded );
    fclose(f);
    return false;
  }

  for (n=0; n<NInit; n++)
    fscanf(f, "%lf\n", &xyzInit[n][2]);

  for (n=0; n<NInit; n++)
    fscanf(f, "%lf %lf\n", &xyzInit[n][0], &xyzInit[n][1]);

  // if there is any dislocation info, read it
  fscanf(f, "%lf \n %lf \n %lf \n %lf \n", &unused, &period[0], &period[1], &period[2]);
  fscanf(f, "CORE { %lf;%lf }\n", &xCore, &yCore);

  // the Burgers vector is assumed to be parallel to the z-direction
  lenBurgers = period[2];

  // read the rest of the file - information about the relaxation
  pos = ftell(f);
  qtxt = "";

  // is there any reasonable info (not just spaces) ?
  while (!feof(f)) {
    n = fscanf(f,"%s\n",&txt);
    if (n<0)
      break;

    qtxt = QString(txt).remove(' ');
    if (qtxt!="")
      break;
  }

  if (qtxt!="") {
    fseek(f,0,SEEK_END);
    TailSize = ftell(f)-pos+1;
    fseek(f,pos,SEEK_SET);
    
    tail = (char*) malloc(TailSize);
    if (tail==NULL) {
      msgCritical(crit_OutOfMemory);
      exit(2);
    }
  
    fread(tail,1,TailSize,f);
  }

  fclose(f);

  return true;
}


// Calculation of the atomic displacements
void PltWin::PrepareBlock()
{ 
  QString stat;
  double zcoord[MAX_ATOMS][2];
  double zlast, dmin, dmax, disp;
  int i, m, n, d, layer, fgidx, bgidx, thidx, diaidx;


  // calculation of the boundaries of the block of atoms

  for (d=0; d<3; d++) {
    xyzMin[d] = 32767;
    xyzMax[d] = -32767;
  }

  for (n=0; n<NInit; n++) {
    for (d=0; d<3; d++) {
      if (xyzInit[n][d] < xyzMin[d])  xyzMin[d] = xyzInit[n][d];
      if (xyzInit[n][d] > xyzMax[d])  xyzMax[d] = xyzInit[n][d];
    }
  }

  for (d=0; d<3; d++) {
    blSize[d] = xyzMax[d] - xyzMin[d];
    xyzCen[d] = (int) trunc((xyzMin[d]+xyzMax[d])/2.0);
  }

  if (VERBOSE) {
    printf("* number of atoms in the block = %6d\n", NInit);
    printf("* geometry of the block of atoms (unrelaxed coordinates):\n");
    printf("    xmin = %8.4lf,  xmax = %8.4lf,  xsize = %8.4lf\n", 
	   xyzMin[0], xyzMax[0], blSize[0]);
    printf("    ymin = %8.4lf,  ymax = %8.4lf,  ysize = %8.4lf\n", 
	   xyzMin[1], xyzMax[1], blSize[1]);
    printf("    zmin = %8.4lf,  zmax = %8.4lf,  zsize = %8.4lf\n", 
	   xyzMin[2], xyzMax[2], blSize[2]);    
  }


  // find z-layers by sorting the z-coordinates in ascending order

  for (n=0; n<NInit; n++) {
    zcoord[n][0] = n;
    zcoord[n][1] = xyzInit[n][2];
  }

  for (m=0; m<NInit; m++) {
    for (n=0; n<NInit; n++) {
      if (isequal(zcoord[m][1],zcoord[n][1])) {
	zcoord[m][1] = round(zcoord[m][1]*1e4)/1e4;
	zcoord[n][1] = round(zcoord[n][1]*1e4)/1e4;
	continue;
      }

      if (zcoord[m][1]<zcoord[n][1]) {
        swap(zcoord[m][0],zcoord[n][0]);
        swap(zcoord[m][1],zcoord[n][1]);
      }  
    }
  }


  // layers and their appearance

  layer = -1;
  for (n=0; n<NInit; n++) {
    if (n==0 || zcoord[n][1]>zlast) {
      layer++;     
      zlast = zcoord[n][1];

      zCoordLayer[layer] = zlast;

      // calculation of the color, thickness and diameter
      fgidx = CalcIndex(layer,NUM_FGCOLORS);
      bgidx = CalcIndex(layer,NUM_BGCOLORS);
      thidx = CalcIndex(layer,NUM_THICKNESS);
      diaidx = CalcIndex(layer,NUM_DIAM);

      zColorLayer[layer][0] = ATOM_FGCOLOR[fgidx];
      zColorLayer[layer][1] = ATOM_BGCOLOR[bgidx];
      zLineThickLayer[layer] = ATOM_THICKNESS[thidx];
      zDiamLayer[layer] = ATOM_DIAM[diaidx];
    }
    zLayer[(int) zcoord[n][0]] = layer;
  } 

  NumZLayers = layer+1;


  // all layers are selected by default -> i.e. atoms in all layers are considered
  // for plotting arrows

  memset( zLayerSel, 1, NumZLayers );

  if (VERBOSE) {
    printf("* list of z-layers, their positions along the z-axis and visibility:\n");
    printf("  (differential displacements will be calculated only between atoms in visible layers)\n");

    for (i=0; i<NumZLayers; i++) {
      if (zLayerSel[i])
	stat = "visible";
      else
	stat = "hidden";

      printf("    %3d  %8.4lf  %s\n", 
	     i+1, zCoordLayer[i], stat.ascii());
    }
  }


  // calculation of the displacement vectors

  dmin = INF;
  dmax = -INF;

  for (n=0; n<NRel; n++) {
    for (d=0; d<3; d++)
      aDisp[n][d] = xyzRel[n][d] - xyzInit[n][d];    

    if (VERBOSE) {
      disp = sqrt( pow(aDisp[n][0],2) + pow(aDisp[n][1],2) + pow(aDisp[n][2],2) );
      if (disp<dmin)  dmin = disp;
      if (disp>dmax)  dmax = disp;
    }
  }

  if (VERBOSE) {
    printf("* calculation of displacements induced by the applied stress:\n"); 
    printf("    dmin = %8.4lf,  dmax = %8.4lf\n", dmin, dmax);
  }


  // Construct the neighbor list in the unrelaxed configuration
  BuildNeighborList( this, NeighListUnrel );
}
