#include "ddplot.h"
#include "messages.h"
#include "neighlist.h"
#include "pltwin.h"



double rdf_dist[MAX_NEIGHBORS];  // radial distribution function (used in qsort)
int rdf_num[MAX_NEIGHBORS];


/*-------------------------------------------------------------------------
!
!  This function compares two distances in the RDF (for the use in
!  qsort below.
!
!------------------------------------------------------------------------*/

int RdfCompare(const void *idx1, const void *idx2)
{
   if (rdf_dist[*(int *)idx1] < rdf_dist[*(int *)idx2]) 
     return -1;
   else if (rdf_dist[*(int *)idx1] == rdf_dist[*(int *)idx2]) 
     return 0;
   else
     return 1;
}


/*-------------------------------------------------------------------------
!
!  Calculate the radial distribution function - i.e. find the number
!  of atoms which have a certain distance from a particular central
!  atom.
!
!------------------------------------------------------------------------*/

void CalcRDF( PltWin *pw )
{
  double dx, dy, dz, dist;
  int icell[3], idxs[MAX_NEIGHBORS], icmin[3], icmax[3];
  int i, d, idx, icen, ineigh, iidx, jidx, kidx;
  bool written;


  pw->RdfNNeigh = 0;

  for (icen=0; icen<pw->NInit; icen++) {
    // consider only atoms in the selected Z-layers
    if (!pw->zLayerSel[ pw->zLayer[icen] ])
      continue;

    // get the indices of the cell in which atom icen resides
    GetCell(pw, icen, icell);

    // we will search also the neighboring cells - get the indices
    for (d=0; d<3; d++) {
      icmin[d] = (icell[d]-1 >= 0)  ?  icell[d]-1 : 0;  
      icmax[d] = (icell[d]+1 <= pw->ncell[d]-1)  ?  icell[d]+1 : pw->ncell[d]-1;
    }


    // search through all neighboring cells and itself
    for (iidx=icmin[0]; iidx<=icmax[0]; iidx++) {
      for (jidx=icmin[1]; jidx<=icmax[1]; jidx++) {
	for (kidx=icmin[2]; kidx<=icmax[2]; kidx++) {

	  // get the head atom (one of possible neighbors of icen) for this cell
	  ineigh = pw->head[iidx][jidx][kidx];

	  while (ineigh>=0) {
	    // skip itself
	    if (icen==ineigh) {
	      ineigh = pw->list[ineigh];
	      continue;
	    }

	    // consider only the neighboring atoms located in selected Z-layers
	    if (!pw->zLayerSel[ pw->zLayer[ineigh] ]) {
	      ineigh = pw->list[ineigh];
	      continue;
	    }
	    
	    dx = pw->xyzInit[ineigh][0] - pw->xyzInit[icen][0];
	    dy = pw->xyzInit[ineigh][1] - pw->xyzInit[icen][1]; 
	    dz = pw->xyzInit[ineigh][2] - pw->xyzInit[icen][2]; 
	    
	    // calculate the distance between icen and ineigh
	    dist = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2));
	    written = false;
	    
	    if (dist<=pw->rcut) {
	      // add into the array of atomic separations if it fits in some
	      for (i=0; i<pw->RdfNNeigh; i++) {
		if (dist <= pw->RdfDist[i]+1e-3 && dist >= pw->RdfDist[i]-1e-3) {
		  pw->RdfNum[i]++;
		  written = true;
		  break;
		}
	      }
	      	      
	      // a completely new interatomic distance
	      if (!written) {
		pw->RdfNum[ pw->RdfNNeigh ] = 1;
		pw->RdfDist[ pw->RdfNNeigh ] = dist;
		pw->RdfNNeigh++;
		
		if (pw->RdfNNeigh >= MAX_NEIGHBORS) {
		  msgCritical( crit_MAX_NEIGHBORS_Exceeded );
		  icen = pw->NInit;
		  ineigh = -1;
		  
		  exit(0);
		}
	      }
	    }

	    // take next atom in the same cell (-1 of no other exists)
	    ineigh = pw->list[ineigh];
	  };
	}
      }
    }
  }

  // generate indexes into RdfDist and RdfNum
  for (i=0; i<pw->RdfNNeigh; i++)
    idxs[i] = i;
 
  // sort according to RdfDist in the ascending order -> after
  // sorting the qsort returns indexes <idxs> such that the array
  // RdfDist[idxs[:]] is sorted
  memcpy(rdf_dist, pw->RdfDist, sizeof(pw->RdfDist));
  memcpy(rdf_num, pw->RdfNum, sizeof(pw->RdfNum));

  qsort(idxs, pw->RdfNNeigh, sizeof(int), RdfCompare);

  // save the sorted arrays
  for (i=0; i<pw->RdfNNeigh; i++) {
    pw->RdfDist[i] = rdf_dist[ idxs[i] ];
    pw->RdfNum[i] = rdf_num[ idxs[i] ];
  }
}
